import pandas as pd
import matplotlib.pyplot as plt
from scipy import linalg
import matplotlib as mpl
import math
import numpy as np
import re as reg
import datetime as dt
from collections import defaultdict
from scipy.spatial import distance
from scipy.cluster import hierarchy
import sklearn.preprocessing as preprocess
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from sklearn.metrics.pairwise import euclidean_distances
from sklearn.neighbors import DistanceMetric, KNeighborsClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import LinearSVC, SVC, SVR, LinearSVR
from sklearn.linear_model import LogisticRegression, ElasticNet, ElasticNetCV, LinearRegression
from sklearn.metrics import brier_score_loss, precision_score, recall_score, f1_score, accuracy_score, \
    mean_absolute_error, mean_squared_error, mean_squared_log_error, r2_score, roc_auc_score, roc_curve, recall_score, \
    confusion_matrix, explained_variance_score, mean_absolute_error, mean_squared_error
from sklearn.calibration import CalibratedClassifierCV, calibration_curve
from sklearn.model_selection import train_test_split, learning_curve, ShuffleSplit, cross_val_score, cross_val_predict
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF
from sklearn import svm
from sklearn.covariance import EllipticEnvelope
from sklearn.ensemble import IsolationForest, RandomForestClassifier, RandomForestRegressor, GradientBoostingClassifier, \
    GradientBoostingRegressor
from sklearn.neighbors import LocalOutlierFactor
from scipy import stats
from sklearn.cluster import AgglomerativeClustering
from sklearn.cluster import DBSCAN
from sklearn import mixture
import itertools
import random
from sklearn.externals import joblib
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import learning_curve
from sklearn.kernel_ridge import KernelRidge
import os
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import average_precision_score
from sklearn.metrics import roc_auc_score
import time
import argparse

color_iter = itertools.cycle(['b', 'g', 'r', 'c', 'm', 'y', 'k', 'grey', 'saddlebrown', 'orchid'])
colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'grey', 'saddlebrown', 'orchid']

# ----------------------------------------------------------------------
def writeModelinFile(model_id, model_name, desc_ident, model, dirFA):
    filename = model_name + '_' + str(model_id) + desc_ident + '.pkl'
    filenamefull = dirFA + '/'+ filename
    df = pd.DataFrame([filename])
    file_list = dirFA + '/list.csv'
    with open(file_list, 'a') as f:
        df.to_csv(f, encoding='utf-8', header=True)

    joblib.dump(model, filenamefull)

# ----------------------------------------------------------------------
class ReturnClassificatorValues(object):
    def __init__(self, accuracy, specificity, sensitivity, auc, mcc, prc, roc):
        self.accuracy = accuracy
        self.specificity = specificity
        self.sensitivity = sensitivity
        self.auc = auc
        self.mcc = mcc
        self.prc = prc
        self.roc = roc


# ----------------------------------------------------------------------
def classificatorTraining(model, a, i, estimator_name, test, test_y, dirFA):
    y_pred = model.predict(test)
    accuracy = accuracy_score(test_y, y_pred)
    confusion_mat = confusion_matrix(test_y, y_pred)
    TN = confusion_mat[0, 0]
    FP = confusion_mat[0, 1]
    sensitivity = recall_score(test_y, y_pred)
    all_negative = TN + FP
    specificity = TN / all_negative
    y_pred_prob = model.predict_proba(test)[:, 1]
    fpr_prob, tpr_prob, _ = roc_curve(test_y, y_pred_prob)
    aucp = roc_auc_score(test_y, y_pred_prob)
    mcc = matthews_corrcoef(test_y, y_pred)
    prc = average_precision_score(test_y, y_pred)
    roc = roc_auc_score(test_y, y_pred)
    title_prob = 'ROC_curve_prob_' + estimator_name
    fig = plt.figure(2)
    plt.plot([0, 1], [0, 1], 'k--')
    plt.plot(fpr_prob, tpr_prob,
             label='ROC curve (area = %0.2f)' % (aucp))
    plt.xlabel('False positive rate')
    plt.ylabel('True positive rate')
    plt.title(title_prob)
    plt.legend(loc='best')
    plot_filename_prob = dirFA + '/'+ title_prob + '_' + str(a) + '_' + str(i) + '.jpg'
    plt.savefig(plot_filename_prob)
    plt.close(fig)

    fpr, tpr, _ = roc_curve(test_y, y_pred, pos_label=1)
    auc = roc_auc_score(test_y, y_pred)
    title = 'ROC_curve_' + estimator_name
    fig2 = plt.figure(3)
    plt.plot([0, 1], [0, 1], 'k--')
    plt.plot(fpr, tpr,
             label='ROC curve (area = %0.2f)' % (auc))
    plt.xlabel('False positive rate')
    plt.ylabel('True positive rate')
    plt.title(title)
    plt.legend(loc='best')
    plot_filename = dirFA + '/'+title + '_' + str(a) + '_' + str(i) + '.jpg'
    plt.savefig(plot_filename)
    plt.close(fig2)
    return ReturnClassificatorValues(accuracy, specificity, sensitivity, auc, mcc, prc, roc)


# ----------------------------------------------------------------------
def classificatorFinalTraining(model, mod_name, x, y, a, dirFA):
    y_pred = model.predict(x)
    accuracy = accuracy_score(y, y_pred)
    confusion_mat = confusion_matrix(y, y_pred)
    TN = confusion_mat[0, 0]
    FP = confusion_mat[0, 1]
    sensitivity = recall_score(y, y_pred)
    all_negative = TN + FP
    specificity = TN / all_negative
    y_pred_prob = model.predict_proba(x)[:, 1]
    fpr_prob, tpr_prob, _ = roc_curve(y, y_pred_prob)
    aucp = roc_auc_score(y, y_pred_prob)
    mcc = matthews_corrcoef(y, y_pred)
    prc = average_precision_score(y, y_pred)
    roc = roc_auc_score(y, y_pred)
    title_prob = 'ROC_curve_prob_' + mod_name
    fig = plt.figure(2)
    plt.plot([0, 1], [0, 1], 'k--')
    plt.plot(fpr_prob, tpr_prob,
             label='ROC curve (area = %0.2f, acc = %0.2f, spec = %0.2f, sen = %0.2f)' % (
             aucp, accuracy, specificity, sensitivity))
    plt.xlabel('False positive rate')
    plt.ylabel('True positive rate')
    plt.title(title_prob)
    plt.legend(loc='best')
    plot_filename_prob = dirFA + '/'+ title_prob + '_' + str(a) + '.jpg'
    plt.savefig(plot_filename_prob)
    plt.close(fig)

    fpr, tpr, _ = roc_curve(y, y_pred, pos_label=1)
    auc = roc_auc_score(y, y_pred)
    title = 'ROC_curve_' + mod_name
    fig2 = plt.figure(3)
    plt.plot([0, 1], [0, 1], 'k--')
    plt.plot(fpr, tpr,
             label='ROC curve (area = %0.2f, acc = %0.2f, spec = %0.2f, sen = %0.2f)' % (
             auc, accuracy, specificity, sensitivity))
    plt.xlabel('False positive rate')
    plt.ylabel('True positive rate')
    plt.title(title)
    plt.legend(loc='best')
    plot_filename = dirFA + '/' + title + '_' + str(a) + '.jpg'
    plt.savefig(plot_filename)
    plt.close(fig2)
    return ReturnClassificatorValues(accuracy, specificity, sensitivity, auc, mcc, prc, roc)

# ----------------------------------------------------------------------
def models_preparation_train(desc_dist, model, dirFA, dirROC, dirModel):

    estimator = LogisticRegression()

    file_groups = dirFA + '/groups_size.csv'
    dfGroups = pd.read_csv(file_groups)
    groups_amount = 0
    cv_groups = 0
    for ind_gr, row_gr in dfGroups.iterrows():
        groups_amount = row_gr['groups_amount']
        cv_groups = row_gr['cv_groups']
    k = 0


    if model == 'Logistic_Regression':
        estimator = LogisticRegression(C=1e5)
    if model == 'RandomForestClassifier':
        estimator = RandomForestClassifier()
    if model == 'RandomForestClassifier2':
        estimator = RandomForestClassifier(max_depth=5, n_estimators=10, max_features=1)
    if model == 'GradientBoostingClassifier':
        estimator = GradientBoostingClassifier()
    if model == 'SVC_Linear':
        estimator =GridSearchCV(SVC(kernel='linear', probability=True), cv=cv_groups,
                           param_grid={"C": [1e0, 1e1, 1e2, 1e3], "gamma": np.logspace(-2, 2, 5)})
    if model == 'SVC_RBF':
        estimator =GridSearchCV(SVC(kernel='rbf', gamma=0.1, probability=True), cv=cv_groups,
                              param_grid={"C": [1e0, 1e1, 1e2, 1e3], "gamma": np.logspace(-2, 2, 5)})

    for x in range(int(groups_amount)):
        model_x_set = pd.DataFrame()
        model_y_set = pd.DataFrame()
        dfPredClas = pd.DataFrame()
        filename_clas = dirFA + '/filename_clas_desc_' + model+ '.csv'
        for a in range(int(cv_groups)):
            filename_train = dirFA + '/'+ str(x) + '_' + str(a) + '_train.csv'
            new_train_df = pd.read_csv(filename_train, encoding='utf-8')
            filename_test = dirFA + '/'+ desc_dist + '_' + str(x) + '_' + str(a) + '_test.csv'
            new_test_df = pd.read_csv(filename_test, encoding='utf-8')

            test_df_y = new_test_df['Class']
            test_df_x = new_test_df.drop(['Unnamed: 0', 'Freq', 'Met', 'Broken', 'Class'], 1)

            train_df_y = new_train_df['Class']
            train_df_x = new_train_df.drop(['Unnamed: 0', 'Freq', 'Met', 'Broken', 'Class'], 1)

            if a == 0:
                model_x_set = train_df_x
                model_x_set = model_x_set.append(test_df_x)
                model_y_set = train_df_y
                model_y_set = model_y_set.append(test_df_y)

            interm_Model = estimator.fit(train_df_x, train_df_y)
            trained_Model = classificatorTraining(interm_Model, x, a, model, test_df_x, test_df_y, dirROC)

            acc = 'accuracy_'+model
            sen = 'sensitivity_' + model
            spec = 'specificity_' + model
            mcc = 'mcc_' + model
            prc = 'prc_' + model
            roc = 'roc_' + model
            auc = 'auc_' + model
            dfPredictionClas = pd.DataFrame(
                {acc: trained_Model.accuracy, sen: trained_Model.sensitivity,
                 spec: trained_Model.specificity, mcc: trained_Model.mcc,
                 prc: trained_Model.prc,roc: trained_Model.roc,auc:trained_Model.auc}, index=[k])
            dfPredClas = dfPredClas.append(dfPredictionClas)
        k = k + 1

        final_Model = estimator.fit(model_x_set, model_y_set)
        classificatorFinalTraining(final_Model, model, model_x_set, model_y_set, x, dirROC)

        dfPredClas.loc["MEAN"] = dfPredClas.mean(axis=0)
        with open(filename_clas, 'a') as f:
            dfPredClas.to_csv(f, encoding='utf-8', header=True)

        writeModelinFile(x, model, 'desc', final_Model, dirModel)

# ----------------------------------------------------------------------
def models_preparation_dts(dataFA, desc_dist, dirFA):
    meantrain = dataFA.mean()
    dataFA = dataFA.fillna(meantrain)
    dataFA_Desc_forModel2 = dataFA

    nonzeros = dataFA_Desc_forModel2['Broken'].astype('int64', copy=False) != 0
    nonzeros_df = dataFA_Desc_forModel2[nonzeros]
    inds_nonzeros_list = nonzeros_df.index.tolist()
    random.shuffle(inds_nonzeros_list)
    inds_nonzeros = set(inds_nonzeros_list)

    cv_groups = 5
    res_div = int(float(nonzeros_df.shape[0]) / 5.0)
    perc = 4.0 / 5
    if res_div < 2:
        res_div = int(float(nonzeros_df.shape[0]) / 4.0)
        if res_div >= 2:
            perc = 3.0 / 4
            cv_groups = 4
        else:
            res_div = int(float(nonzeros_df.shape[0]) / 3.0)
            if res_div >= 2:
                perc = 2.0 / 3
                cv_groups = 3
            else:
                res_div = int(float(nonzeros_df.shape[0]) / 2.0)
                if res_div >= 2:
                    perc = 1.0 / 2
                    cv_groups = 2
                else:
                    cv_groups = 1
                    perc = 1.0

    nonzero_train = int(nonzeros_df.shape[0] * perc + 0.5)

    nonzero_test = nonzeros_df.shape[0] - nonzero_train

    checked = []
    zeros = dataFA_Desc_forModel2['Broken'].astype('int64', copy=False) == 0
    zeros_df = dataFA_Desc_forModel2[zeros]
    indeces_zeros_list = zeros_df.index.tolist()
    inds_zeros_s = set(indeces_zeros_list)

    groups_amount = round(zeros_df.shape[0] / nonzeros_df.shape[0])

    inds_zeros_sels = defaultdict(list)

    dfGroups = pd.DataFrame({'groups_amount': groups_amount, 'cv_groups': cv_groups, }, index=[0])
    filename_groups = dirFA+ '/' + 'groups_size.csv'
    dfGroups.to_csv(filename_groups, encoding='utf-8')

    for y in range(groups_amount):
        if y == 0:
            for_selection_l = indeces_zeros_list
            random.shuffle(for_selection_l)
            inds_zeros_sel = random.sample(for_selection_l, nonzeros_df.shape[0])
            checked = checked + inds_zeros_sel
            inds_zeros_sels[y] = inds_zeros_sel
        elif 0 < y < (groups_amount - 1):
            checked_s = set(checked)
            for_selection = inds_zeros_s.difference(checked_s)
            for_selection_l = list(for_selection)
            random.shuffle(for_selection_l)
            inds_zeros_sel = random.sample(for_selection_l, nonzeros_df.shape[0])
            checked = checked + inds_zeros_sel
            inds_zeros_sels[y] = inds_zeros_sel
        else:
            checked_s = set(checked)
            for_selection = inds_zeros_s.difference(checked_s)
            for_selection_l = list(for_selection)
            size_add = nonzeros_df.shape[0] - len(for_selection_l)
            random.shuffle(for_selection_l)
            if size_add > 0:
                inds_zeros_add = random.sample(checked, size_add)
                inds_zeros_sel = for_selection_l + inds_zeros_add
                inds_zeros_sels[y] = inds_zeros_sel
            else:
                inds_zeros_sels[y] = for_selection_l

    a = 0
    for x in range(groups_amount):
        inds_zeros_sel = inds_zeros_sels[x]
        inds_zeros_sel_s = set(inds_zeros_sel)
        inds_train_zero_dict = defaultdict(list)
        inds_test_zero_dict = defaultdict(list)
        inds_train_nonzero_dict = defaultdict(list)
        inds_test_nonzero_dict = defaultdict(list)

        for i in range(cv_groups):
            st_pos_z = nonzero_test * i
            end_pos_z = nonzero_test * (i + 1)
            inds_nonzeros_test = inds_nonzeros_list[st_pos_z:end_pos_z]
            inds_nonzeros_test_s = set(inds_nonzeros_test)
            inds_nonzeros_train_s = inds_nonzeros.difference(inds_nonzeros_test_s)
            inds_nonzeros_train = list(inds_nonzeros_train_s)
            inds_zeros_test = inds_zeros_sel[st_pos_z:end_pos_z]
            inds_zeros_test_s = set(inds_zeros_test)
            inds_zeros_train_s = inds_zeros_sel_s.difference(inds_zeros_test_s)
            inds_zeros_train = list(inds_zeros_train_s)
            inds_train_zero_dict[i] = inds_zeros_train
            inds_test_zero_dict[i] = inds_zeros_test
            inds_train_nonzero_dict[i] = inds_nonzeros_train
            inds_test_nonzero_dict[i] = inds_nonzeros_test

        for y in range(cv_groups):
            desc_train_0_x = dataFA_Desc_forModel2.iloc[inds_train_zero_dict.get(y)]
            desc_test_0_x = dataFA_Desc_forModel2.iloc[inds_test_zero_dict.get(y)]

            train_df = dataFA_Desc_forModel2.iloc[inds_train_nonzero_dict.get(y)]
            test_df = dataFA_Desc_forModel2.iloc[inds_test_nonzero_dict.get(y)]

            zeroslist_train = np.zeros(desc_train_0_x.shape[0])
            zeroslist_test = np.zeros(desc_test_0_x.shape[0])
            oneslist_train = np.ones(train_df.shape[0])
            oneslist_test = np.ones(test_df.shape[0])
            train_df['Class'] = pd.Series(oneslist_train, index=train_df.index)
            test_df['Class'] = pd.Series(oneslist_test, index=test_df.index)
            desc_train_0_x['Class'] = pd.Series(zeroslist_train, index=desc_train_0_x.index)
            desc_test_0_x['Class'] = pd.Series(zeroslist_test, index=desc_test_0_x.index)

            train_df = train_df.append(desc_train_0_x)
            test_df = test_df.append(desc_test_0_x)

            new_train_df = pd.DataFrame()
            new_test_df = pd.DataFrame()

            mets_train = defaultdict(list)
            mets_test = defaultdict(list)

            for row, value in train_df.iterrows():
                for idx, x in value.items():
                    if idx == 'Met':
                        int_x = int(x)
                        mm = int_x * 11 / (100 + int_x)
                        mets_train[int_x] = int(mm)

            for rowt, valuet in test_df.iterrows():
                for idx, x in valuet.items():
                    if idx == 'Met':
                        int_x = int(x)
                        mm = int_x * 11 / (100 + int_x)
                        mets_test[int_x] = int(mm)

            for am in mets_train.keys():
                rows = train_df['Met'] == am
                mm = mets_train[am]
                df_try = train_df[rows]
                int_mm = int(mm) + 1
                new_df = [df_try] * int_mm
                new_train_df = new_train_df.append(new_df, ignore_index=True)

            for am in mets_test.keys():
                rows = test_df['Met'] == am
                mm = mets_test[am]
                int_mm = int(mm) + 1
                df_try = test_df[rows]
                new_df = [df_try] * int_mm
                new_test_df = new_test_df.append(new_df, ignore_index=True)

            filename_train = dirFA + '/'+str(int(a)) + '_' + str(int(y)) + '_train.csv'
            new_train_df.to_csv(filename_train, encoding='utf-8')
            filename_test = dirFA+ '/' + desc_dist + '_' + str(int(a)) + '_' + str(int(y)) + '_test.csv'
            new_test_df.to_csv(filename_test, encoding='utf-8')
            df_test = pd.DataFrame([filename_test])
            df_train = pd.DataFrame([filename_train])
        a = a + 1
    file_list_tr = dirFA+ '/' + 'list_train.csv'
    with open(file_list_tr, 'a') as f:
        df_train.to_csv(f, encoding='utf-8', header=True)

    file_list_ts = dirFA+ '/' + 'list_test.csv'
    with open(file_list_ts, 'a') as f:
        df_test.to_csv(f, encoding='utf-8', header=True)
 # ----------------------------------------------------------------------

if __name__ == '__main__':

    # # Parse command line arguments
    parser = argparse.ArgumentParser(prog='Description of the program')
    #

    #-wd C:/Users/Tatiana/Desktop/new_models/ -rd C:/Users/Tatiana/Desktop/test_models/ -pf proteases.csv -df scaled_descriptorsVS.csv -mf models.csv -n 4 -c 4 -s 1

    # Define all arguments
    parser.add_argument('-wd', '--working_directory', required=True, dest='working_directory', help='Working_directory where input files are stored')
    parser.add_argument('-pf', '--protease_file', required=True, dest='protease_file', help='File with protease names')
    parser.add_argument('-df', '--descriptors_file', required=True, dest='descriptors_file', help='File with descriptors')
    parser.add_argument('-mf', '--models_file', required=True, dest='models_file', help='File with models names')
    parser.add_argument('-n', '--n_residues', required=True, dest='n_residues', help='Amount of residues to be considered in n-terminal direction from cleavage site (p1, p2 etc), where cleavage site is between P1 and P`1')
    parser.add_argument('-c', '--с_residues', required=True, dest='c_residues', help='Amount of residues to be considered in с-terminal direction from cleavage site (p`1, p`2 etc), where cleavage site is between P1 and P`1')
    parser.add_argument('-s', '--stage', required=True, dest='stage', help='Stage of code. 1 - Prepare_main_dataset. 2 - Prepare_train_test_datasets. 3- Build machine learning prediction model.')

    # Input file should be .csv
    # Directory should end with /
    # Descriptors type is either shop or volsurf
    # Parse arguments. ###INPUT PARAMETER
    args = parser.parse_args()
    dirWork = args.working_directory
    protease_csv = args.protease_file
    descr_file = args.descriptors_file
    models_file = args.models_file
    amount_N_terminal = int(args.n_residues)
    amount_C_terminal = int(args.c_residues)
    stage = args.stage
    site_of_cleavage_size = amount_N_terminal + amount_C_terminal


    # Read csv file with list of proteases for model building
    ADV_Proteases = dirWork+'/' +protease_csv
    dfProteases = pd.read_csv(ADV_Proteases)

    # Read csv file with list of learning algorithms for model building
    ADV_models = dirWork+'/' + models_file
    dfModels = pd.read_csv(ADV_models)

    # Read csv file with Volsurf Descriptors that characterise each of the structural blocks in the database for model building
    ADV_DESC = dirWork+'/' + descr_file
    dfVSDescriptors = pd.read_csv(ADV_DESC, sep='\t',encoding='utf-8')

    # Define the stage to be completed during the model building
    step_to_perform = ''
    if stage == '1':
        step_to_perform = 'Prepare_main_dataset'
    elif stage == '2':
        step_to_perform = 'Prepare_train_test_datasets'
    elif stage == '3':
         step_to_perform = 'Train_models'

    for ind_pr, row_pr in dfProteases.iterrows():

        dataFA_Desc_forModel = pd.DataFrame()

        protease = row_pr['name']

        dirProtSize = dirWork + '/'+ protease + '/'+ str(site_of_cleavage_size)

        file_namedesc =  dirProtSize+ '/'+ protease + '_descriptors_main_dataset.csv'

        # Prepare main dataset for each of the selected proteases. Here we cut out all possible n-size SBs sequences from substrate peptide.
        # N-size is defined by amount_N_terminal - amount of residues to be considered in direction of N-terminal of the substrate peptide and amount_C_terminal - amount of residues to be considered in direction of C-terminal of the substrate peptide.

        if step_to_perform == 'Prepare_main_dataset':

            ADVFAPEPT =  dirProtSize +'/dataset.csv'
            dataFAPept = pd.read_csv(ADVFAPEPT, sep='\t', encoding='utf-8')

            mets = dataFAPept['amount'].astype('int64', copy=False)
            broks = dataFAPept['frequency'].astype('int64', copy=False)
            exp_results = dataFAPept['class'].astype('int64', copy=False)

            freqs = np.array([])

            a = 0

            for ind_F, row_F in dataFAPept.iterrows():
                frames = []
                pr_inchikeys = []
                p_inchikeys = []
                for am in range(amount_N_terminal):
                    index=amount_N_terminal-am
                    p_name = 'p'+str(index)
                    fh_inchikey_p = row_F[p_name]
                    p_inchikeys.append(fh_inchikey_p)
                    descP = dfVSDescriptors.loc[dfVSDescriptors['Fh_IK'] == fh_inchikey_p]
                    descP.index = [a]
                    columnsPnew = np.array([])
                    for ind in range(descP.columns.values.size):
                        column_name_P = descP.columns.values.item(ind)
                        namenew = column_name_P + '_P'+str(index)
                        columnsPnew = np.append(columnsPnew, namenew)
                    descP.columns = columnsPnew
                    frames.append(descP)


                for am in range(amount_C_terminal):
                    index = am+1
                    pr_name = 'pr' + str(index)
                    fh_inchikey_pr = row_F[pr_name]
                    pr_inchikeys.append(fh_inchikey_pr)
                    descPr = dfVSDescriptors.loc[dfVSDescriptors['Fh_IK'] == fh_inchikey_pr]
                    descPr.index = [a]
                    columnsPrnew = np.array([])
                    for ind in range(descPr.columns.values.size):
                        column_name_Pr = descPr.columns.values.item(ind)
                        namenew = column_name_Pr + '_Pr'+str(index)
                        columnsPrnew = np.append(columnsPrnew, namenew)
                    descPr.columns = columnsPrnew
                    frames.append(descPr)

                term_name_p = 'p'+str(amount_N_terminal)+'_term'
                term_p4 = row_F[term_name_p]
                term_name_pr = 'pr' + str(amount_C_terminal) + '_term'
                term_pr4 = row_F[term_name_pr]

                bond_met = row_F['amount']
                fa_met = float(bond_met)
                bond_br = row_F['frequency']
                fa_broken = float(bond_br)
                frequency = float(fa_broken / fa_met)
                freqs = np.append(freqs, frequency)

                descriptors_freq = pd.concat(frames, axis=1)

                dataFA_Desc_forModel = dataFA_Desc_forModel.append(descriptors_freq)

                a = a + 1

            dataFA_Desc_forModel['Freq'] = pd.Series(freqs, index=dataFA_Desc_forModel.index)
            dataFA_Desc_forModel['Met'] = pd.Series(mets, index=dataFA_Desc_forModel.index)
            dataFA_Desc_forModel['Broken'] = pd.Series(broks, index=dataFA_Desc_forModel.index)
            dataFA_Desc_forModel['Class'] = pd.Series(exp_results, index=dataFA_Desc_forModel.index)
            dataFA_Desc_forModel.to_csv(file_namedesc, encoding='utf-8')

        # From unbalanced main dataset create balanced subsets. Each subset is divided in train and test sets. These sets are saved in .csv files.
        elif step_to_perform == 'Prepare_train_test_datasets':

            dataFA_Desc_forModel = pd.read_csv(file_namedesc)
            columns = ''
            frames = []
            for am in range(amount_N_terminal):
                index = amount_N_terminal - am
                p_name = 'Fh_IK_P' + str(index)
                p_hname = 'Helm_Name_P' + str(index)
                p_unname = 'Unnamed: 0_P' + str(index)
                frames.append(p_name)
                frames.append(p_hname)
                frames.append(p_unname)

            for am in range(amount_C_terminal):
                index = am + 1
                pr_name = 'Fh_IK_Pr' + str(index)
                pr_hname = 'Helm_Name_Pr' + str(index)
                pr_unname = 'Unnamed: 0_Pr' + str(index)
                frames.append(pr_name)
                frames.append(pr_hname)
                frames.append(pr_unname)

            dataFA_Desc_forModel2 = dataFA_Desc_forModel.drop(frames, 1)
            models_preparation_dts(dataFA_Desc_forModel2, 'Desc', dirProtSize)

        # Prepared train and test sets are used to train models and each model is evaluated with performance metrics during 5-fold CV (accuracy, AUC ROC, AUC PRC, MCC, Sensitivity and Specificity).
        # Results are saved in the output .csv file.
        elif step_to_perform == 'Train_models':
            for indm, rowm in dfModels.iterrows():
                model = rowm['name']
                dirModel = dirProtSize + '/'+ model
                os.mkdir(dirModel)
                dirROC = dirModel + '/ROC'
                os.mkdir(dirROC)
                models_preparation_train('Desc',model, dirRProt, dirROC, dirModel)




